;;;; uuid-lib.scm
;;;; Kon Lovett, May '06
;;;; BSD

;;; Wrapper for Theodore Y. Ts'o OSF DCE 1.1 UUID Suite
;;; http://lib.sourceforge.net/
;;; Part of MacOS X & Linux (every modern unix?)

(declare
  (bound-to-procedure
    ##sys#signal-hook))

(module uuid-lib

(;export
  ;
  uuid? check-uuid error-uuid
  uuid-null?
  uuid-compare
  uuid=? uuid<? uuid>? uuid<=? uuid>=?
  uuid-v4-hash uuid-hash
  ;
  uuid-copy
  uuid-clear!
  make-uuid
  ;
  uuid-generate
  uuid-parse
  uuid-unparse
  ;
  string->uuid uuid->string
  uuid-print)

(import scheme)
(import (chicken base))
(import (chicken type))
(import (chicken fixnum))
(import (chicken format))
(import (chicken bitwise))
(import (chicken condition))
(import (chicken foreign))
(import (chicken memory))
(import (chicken gc))
(import (chicken locative))

;NOTE works w/o recourse to `define-record' for tag, so ok for tag "defining" module
(cond-expand
  ((or chicken-5.0 chicken-5.1)
    (define (set-record-printer! tag proc)
      (##sys#register-record-printer tag proc) ) )
  (else) )

(define-type unboxed-uuid pointer)
(define-type boxed-uuid (struct uuid))

(define-type error-argument-name (or string symbol))

(: uuid? (* -> boolean : boxed-uuid))
(: error-uuid (symbol * #!optional error-argument-name -> boxed-uuid))
(: check-uuid (symbol * #!optional error-argument-name -> boxed-uuid))
(: uuid-null? (* --> boolean)) ;not a `predicate'
(: uuid-v4-hash (boxed-uuid --> fixnum))
(: uuid-hash (boxed-uuid --> fixnum))
(: uuid-compare (boxed-uuid boxed-uuid --> fixnum))
(: uuid=? (boxed-uuid boxed-uuid --> boolean))
(: uuid<? (boxed-uuid boxed-uuid --> boolean))
(: uuid>? (boxed-uuid boxed-uuid --> boolean))
(: uuid<=? (boxed-uuid boxed-uuid --> boolean))
(: uuid>=? (boxed-uuid boxed-uuid --> boolean))
(: uuid-clear! (boxed-uuid -> boxed-uuid))
(: uuid-copy (boxed-uuid --> boxed-uuid))
(: uuid-generate (#!optional (or false symbol) --> boxed-uuid))
(: make-uuid (#!optional (or false symbol) --> boxed-uuid))
(: uuid-parse ((or string symbol) --> boxed-uuid))
(: uuid-unparse (boxed-uuid #!optional (or false symbol) --> string))
(: string->uuid (string --> boxed-uuid))
(: uuid->string (boxed-uuid --> string))

;

(define-inline (fxabs n) (if (fx< n 0) (fxneg n) n))

(define-inline (fxnegative? n) (fx< n 0))
(define-inline (fxzero? n) (fx= n 0))
(define-inline (fxpositive? n) (fx> n 0))

;

(define get-u64
  (foreign-lambda* unsigned-integer64 (((c-pointer unsigned-integer64) u64p))
    "return( *u64p );"))

;srfi-69

;; Force Hash to Bounded Fixnum:
(define-constant HASH-DEFAULT-BOUND 536870912)

;@hsh : unsigned-integer
(define-inline (32bit-hash-limit hsh)
  ;(assert (unsigned? hsh))
  ;; use 32-bit mask to have identical hashes on 64-bit platforms
  (modulo (bitwise-and hsh (foreign-value "C_MOST_POSITIVE_32_BIT_FIXNUM" int))
          HASH-DEFAULT-BOUND) )

;;

#>
#include <uuid/uuid.h>
<#

(define UUID-SIZEOF (foreign-value "sizeof(uuid_t)" int))

(define-constant UUID-STRING-LENGTH 36)

(define uuid_is_null
  (foreign-lambda bool uuid_is_null nonnull-c-pointer))

(define uuid_compare
  (foreign-lambda int uuid_compare nonnull-c-pointer nonnull-c-pointer))

(define uuid_clear
  (foreign-lambda void uuid_clear nonnull-c-pointer))

(define uuid_copy
  (foreign-lambda void uuid_copy nonnull-c-pointer nonnull-c-pointer))

(define uuid_generate
  (foreign-lambda void uuid_generate nonnull-c-pointer))

(define uuid_generate_random
  (foreign-lambda void uuid_generate_random nonnull-c-pointer))

(define uuid_generate_time
  (foreign-lambda void uuid_generate_time nonnull-c-pointer))

(define uuid_parse
  (foreign-lambda int uuid_parse c-string nonnull-c-pointer))

(define uuid_unparse
  (foreign-lambda void uuid_unparse nonnull-c-pointer nonnull-c-pointer))

(define uuid_unparse_upper
  (foreign-lambda void uuid_unparse_upper nonnull-c-pointer nonnull-c-pointer))

(define uuid_unparse_lower
  (foreign-lambda void uuid_unparse_lower nonnull-c-pointer nonnull-c-pointer))

;;

(define-record-type uuid
  (%make-uuid puuid)
  %uuid?
  (puuid %uuid-puuid))

(define-inline (unbox-puuid boxed-puuid)
  (%uuid-puuid boxed-puuid) )

(define-inline (boxed-puuid-compare uuid1 uuid2)
  (uuid_compare (unbox-puuid uuid1) (unbox-puuid uuid2)) )

(define (free-uuid uuid)
  (free (unbox-puuid uuid)) )

(define-inline (box-puuid puuid)
  (set-finalizer! (%make-uuid puuid) free-uuid) )

(define-inline (new-uuid)
  (box-puuid (allocate UUID-SIZEOF)) )

;;

(define (error-bad-argument-type loc obj #!optional msg)
  (##sys#signal-hook #:type-error loc
                     (apply string-append "bad argument type"
                                          (if msg (list " - " msg) '("")))
                     obj) )

(define (error-uuid-variant loc obj)
  (error-bad-argument-type loc obj "invalid variant") )

(define (error-uuid-form loc obj)
  (error-bad-argument-type loc obj "not a string or symbol") )

(define (error-uuid-string-case loc obj)
  (error-bad-argument-type loc obj "invalid case") )

;;

(define (uuid? obj)
  (%uuid? obj) )

;optional @argnam is ignored
(define (error-uuid loc obj . _)
  (error-bad-argument-type loc obj "not a uuid") )

;optional @argnam is ignored
(define (check-uuid loc obj . _)
  (unless (%uuid? obj) (error-uuid loc obj))
  obj )

(define (uuid-v4-hash uuid)
  ;(assert (= 16 UUID-SIZEOF))
  ;"random" already
  (let ((puuid (unbox-puuid (check-uuid 'uuid-hash uuid))))
    (32bit-hash-limit (get-u64 (pointer+ puuid 8))) ) )

(define (uuid-hash uuid)
  ;(assert (= 16 UUID-SIZEOF))
  ;combine all bits (poorly)
  (let ((puuid (unbox-puuid (check-uuid 'uuid-hash uuid))))
    (32bit-hash-limit (+ (get-u64 puuid) (get-u64 (pointer+ puuid 8)))) ) )

(define (uuid-null? obj)
  (uuid_is_null (unbox-puuid (check-uuid 'uuid-null? obj))) )

(define (uuid-compare uuid1 uuid2)
  (let ((cmp (boxed-puuid-compare
              (check-uuid 'uuid-compare uuid1)
              (check-uuid 'uuid-compare uuid2))) )
    (cond ((negative? cmp) -1)
          ((zero? cmp) 0)
          (else 1) ) ) )

(define (uuid=? uuid1 uuid2)
  (zero? (boxed-puuid-compare uuid1 uuid2)) )

(define (uuid<? uuid1 uuid2)
  (fxnegative? (boxed-puuid-compare uuid1 uuid2)) )

(define (uuid>? uuid1 uuid2)
  (fxpositive? (boxed-puuid-compare uuid1 uuid2)) )

(define (uuid<=? uuid1 uuid2)
  (let ((comp (boxed-puuid-compare uuid1 uuid2)))
    (or (fxzero? comp) (fxnegative? comp))) )

(define (uuid>=? uuid1 uuid2)
  (let ((comp (boxed-puuid-compare uuid1 uuid2)))
    (or (fxzero? comp) (fxpositive? comp))) )

(define (uuid-clear! uuid)
  (and (uuid_clear (unbox-puuid (check-uuid 'uuid-clear! uuid)))
       uuid) )

(define (uuid-copy old-uuid)
  (check-uuid 'uuid-copy old-uuid)
  (let ((uuid (new-uuid)))
    (uuid_copy (unbox-puuid uuid) (unbox-puuid old-uuid))
    uuid) )

(define (uuid-generate #!optional (variant #f))
  (let ((uuid (new-uuid)))
    (case variant
      ((#f)
        (uuid_generate (unbox-puuid uuid)))
      ((v1 V1 time)
        (uuid_generate_time (unbox-puuid uuid)))
      ;
      ((v4 V4 random)
        (uuid_generate_random (unbox-puuid uuid)))
      (else
        (error-uuid-variant 'uuid-generate variant)))
    uuid ) )

(define (make-uuid . args)
  (if (null? args)
      (uuid-clear! (new-uuid))
      (apply uuid-generate args) ) )

(define (uuid-parse uuid-text)
  (unless (or (string? uuid-text) (symbol? uuid-text))
    (error-uuid-form 'uuid-parse uuid-text))
  (let ((uuid-text (if (symbol? uuid-text) (symbol->string uuid-text) uuid-text))
        (uuid (new-uuid)) )
    (and (zero? (uuid_parse uuid-text (unbox-puuid uuid)))
         uuid) ) )

(define (uuid-unparse uuid #!optional (kase #f))
  (check-uuid 'uuid-unparse uuid)
  ;allocate string buffer, +1 for NUL byte; use string so gc
  (let* ((uuid-text (make-string (add1 UUID-STRING-LENGTH)))
         (puuid-text (make-locative uuid-text)) )
    (case kase
      ((#f)
        (uuid_unparse (unbox-puuid uuid) puuid-text))
      ((upper)
        (uuid_unparse_upper (unbox-puuid uuid) puuid-text))
      ((lower)
        (uuid_unparse_lower (unbox-puuid uuid) puuid-text))
      (else
        (error-uuid-string-case 'uuid-generate kase)) )
    ;remove NUL byte
    (substring uuid-text 0 UUID-STRING-LENGTH)) )

(define (string->uuid str)
  (uuid-parse str) )

(define (uuid->string uuid)
  (uuid-unparse uuid 'lower) )

(define (uuid-print obj #!optional (port (current-output-port)))
  (format port "#<~A ~A>" uuid (uuid-unparse obj)) )

(set-record-printer! uuid uuid-print)

) ;module uuid-lib
